"""Working with distance quantities."""

from abc import ABC, abstractproperty

from .quantity import Quantity, UnitSymbol


class DistanceUnit(UnitSymbol):
    """Symbols for distance units."""

    METER = "m"
    METERS = "m"
    M = "m"

    KILOMETER = "km"
    KILOMETERS = "km"
    KM = "km"

    CENTIMETER = "cm"
    CENTIMETERS = "cm"
    CM = "cm"

    MILLIMETER = "mm"
    MILLIMETERS = "mm"
    MM = "mm"

    MILE = "mi"
    MILES = "mi"
    MI = "mi"

    NAUTICAL_MILE = "NM"
    NAUTICAL_MILES = "NM"
    NM = "NM"

    FOOT = "ft"
    FEET = "ft"
    FT = "ft"

    INCH = "in"
    INCHES = "in"
    IN = "in"

    YARD = "yd"
    YARDS = "yd"
    YD = "yd"

    ASTRONOMICAL_UNIT = "au"
    ASTRONOMICAL_UNITS = "au"
    AU = "au"

    PARSEC = "pc"
    PARSECS = "pc"
    PC = "pc"

    LIGHT_YEAR = "ly"
    LIGHT_YEARS = "ly"
    LY = "ly"


class Distance(Quantity, ABC):
    """Base for all distance unit types."""

    @abstractproperty
    def meters(self):
        """Return the value of this quantity in meters."""

    @property
    def kilometers(self):
        """Return the value of this quantity in kilometers."""
        return self.meters * 0.001

    @property
    def centimeters(self):
        """Return the value of this quantity in centimeters."""
        return self.meters * 100

    @property
    def millimeters(self):
        """Return the value of this quantity in millimeters."""
        return self.meters * 1000

    @property
    def miles(self):
        """Return the value of this quantity in miles."""
        return self.feet / 5280

    @abstractproperty
    def feet(self):
        """Return the value of this quantity in feet."""

    @property
    def inches(self):
        """Return the value of this quantity in inches."""
        return self.feet * 12

    @property
    def yards(self):
        """Return the value of this quantity in yards."""
        return self.feet / 3

    @property
    def astronomical_units(self):
        """Return the value of this quantity in astronomical units."""
        return self.meters / 149597870700

    @property
    def nautical_miles(self):
        """Return the value of this quantity in nautical miles."""
        return self.feet * 0.00016458

    @property
    def light_years(self):
        """Return the value of this quantity in light years."""
        return self.meters / 9460730472580800

    @property
    def parsecs(self):
        """Return the value of this quantity in parsecs."""
        return self.miles / 19173511575400

    def __call__(self, type):  # noqa: C901
        """Convert this Distance quantity to the given type."""
        if type == Meter:
            return Meter(self.meters)

        if type == Kilometer:
            return Kilometer(self.kilometers)

        if type == Centimeter:
            return Centimeter(self.centimeters)

        if type == Millimeter:
            return Millimeter(self.millimeters)

        if type == Mile:
            return Mile(self.miles)

        if type == Foot:
            return Foot(self.feet)

        if type == Inch:
            return Inch(self.inches)

        if type == Yard:
            return Yard(self.yards)

        if type == NauticalMile:
            return NauticalMile(self.nautical_miles)

        if type == AstronomicalUnit:
            return AstronomicalUnit(self.astronomical_units)

        if type == LightYear:
            return LightYear(self.light_years)

        if type == Parsec:
            return Parsec(self.parsecs)

        raise TypeError(f"Cannot convert to {type}")


class Meter(Distance, symbol=DistanceUnit.METER):
    """A representation of a meter."""

    @property
    def meters(self):
        """Return the value of this quantity in meters."""
        return self.value

    @property
    def feet(self):
        """Return the value of this quantity in feet."""
        return self.meters * 3.28084


class Millimeter(Meter, symbol=DistanceUnit.MILLIMETER):
    """A representation of a millimeter."""

    @property
    def meters(self):
        """Return the value of this quantity in meters."""
        return self.value * 0.001


class Centimeter(Meter, symbol=DistanceUnit.CENTIMETER):
    """A representation of a centimeter."""

    @property
    def meters(self):
        """Return the value of this quantity in meters."""
        return self.value * 0.01


class Kilometer(Meter, symbol=DistanceUnit.KILOMETER):
    """A representation of a kilometer."""

    @property
    def meters(self):
        """Return the value of this quantity in meters."""
        return self.value * 1000


class NauticalMile(Meter, symbol=DistanceUnit.NAUTICAL_MILE):
    """A representation of a nautical mile."""

    @property
    def meters(self):
        """Return the value of this quantity in meters."""
        return self.value * 1852


class Foot(Distance, symbol=DistanceUnit.FOOT):
    """A representation of foot measurements."""

    @property
    def meters(self):
        """Return the value of this quantity in meters."""
        return self.feet * 0.3048

    @property
    def feet(self):
        """Return the value of this quantity in feet."""
        return self.value


class Mile(Foot, symbol=DistanceUnit.MILE):
    """A representation of a mile."""

    @property
    def feet(self):
        """Return the value of this quantity in feet."""
        return self.value * 5280


class Yard(Foot, symbol=DistanceUnit.YARD):
    """A representation of a yard."""

    @property
    def feet(self):
        """Return the value of this quantity in feet."""
        return self.value * 3


class Inch(Foot, symbol=DistanceUnit.INCH):
    """A representation of an inch."""

    @property
    def feet(self):
        """Return the value of this quantity in feet."""
        return self.value / 12


class AstronomicalUnit(Meter, symbol=DistanceUnit.ASTRONOMICAL_UNIT):
    """A representation of an astronomical unit."""

    @property
    def meters(self):
        """Return the value of this quantity in meters."""
        return self.value * 149597870700


class LightYear(AstronomicalUnit, symbol=DistanceUnit.LIGHT_YEAR):
    """A representation of a light year."""

    @property
    def meters(self):
        """Return the value of this quantity in meters."""
        return self.value * 9460730472580800


class Parsec(AstronomicalUnit, symbol=DistanceUnit.PARSEC):
    """A representation of a parsec."""

    @property
    def meters(self):
        """Return the value of this quantity in meters."""
        return self.value * 30856775814671900



"""Working with pressure quantities."""

from abc import ABC, abstractproperty

from .quantity import Quantity, UnitSymbol


class PressureUnit(UnitSymbol):
    """Symbols for pressure units."""

    HECTOPASCAL = "hPa"
    HECTOPASCALS = "hPa"
    HPA = "hPa"

    PASCAL = "Pa"
    PASCALS = "Pa"
    PA = "Pa"

    INCHES_MERCURY = "inHg"
    INHG = "inHg"

    POUNDS_PER_SQUARE_INCH = "psi"
    LBS_PER_SQ_IN = "psi"
    PSI = "psi"

    BAR = "bar"

    ATMOSPHERE = "atm"
    ATM = "atm"


class Pressure(Quantity, ABC):
    """Base for all pressure unit types."""

    @property
    def bar(self):
        """Return the value of this quantity as bar."""
        return self.hectopascals * 0.001

    @property
    def atmosphere(self):
        """Return the value of this quantity as Atmospheres."""
        return self.pascals / 101325

    @property
    def hectopascals(self):
        """Return the value of this quantity as Hectopascals."""
        return self.pascals * 0.01

    @abstractproperty
    def pascals(self):
        """Return the value of this quantity as Pascals."""

    @abstractproperty
    def pounds_per_sq_in(self):
        """Return the value of this quantity as pounds-per-square-inch."""

    @property
    def inches_mercury(self):
        """Return the value of this quantity as inches-mercury."""
        return self.pounds_per_sq_in * 2.0360206507731

    def __call__(self, type):  # noqa: C901
        """Convert this Pressure quantity to the given type."""
        if type == Pascal:
            return Pascal(self.pascals)

        if type == Hectopascal:
            return Hectopascal(self.hectopascals)

        if type == InchesMercury:
            return InchesMercury(self.inches_mercury)

        raise TypeError(f"Cannot convert to {type}")


class Bar(Pressure, symbol=PressureUnit.BAR):
    """A representation of Bars."""

    @property
    def pascals(self):
        """Return the value of this quantity as Pascals."""
        return self.value * 100000

    @property
    def pounds_per_sq_in(self):
        """Return the value of this quantity as pounds-per-square-inch."""
        return self.value * 14.503773800722


class Atmosphere(Pressure, symbol=PressureUnit.ATMOSPHERE):
    """A representation of Atmospheres."""

    @property
    def pascals(self):
        """Return the value of this quantity as Pascals."""
        return self.value * 101325

    @property
    def pounds_per_sq_in(self):
        """Return the value of this quantity as pounds-per-square-inch."""
        return self.value * 14.695948775513


class Pascal(Pressure, symbol=PressureUnit.PASCAL):
    """A representation of Pascals."""

    @property
    def pascals(self):
        """Return the value of this quantity as Pascals."""
        return self.value

    @property
    def pounds_per_sq_in(self):
        """Return the value of this quantity as pounds-per-square-inch."""
        return self.pascals / 6894.75729


class Hectopascal(Pascal, symbol=PressureUnit.HECTOPASCAL):
    """A representation of Hectopascals."""

    @property
    def pascals(self):
        """Return the value of this quantity as Pascals."""
        return self.value * 100


class PoundsPerSquareInch(Pressure, symbol=PressureUnit.POUNDS_PER_SQUARE_INCH):
    """A representation of PoundsPerSquareInch."""

    @property
    def pascals(self):
        """Return the value of this quantity as Pascals."""
        return self.pounds_per_sq_in * 6894.75729

    @property
    def pounds_per_sq_in(self):
        """Return the value of this quantity as pounds-per-square-inch."""
        return self.value


class InchesMercury(Pressure, symbol=PressureUnit.INCHES_MERCURY):
    """A representation of InchesMercury."""

    @property
    def pascals(self):
        """Return the value of this quantity as Pascals."""
        return self.inches_mercury * 3386.3886666667

    @property
    def inches_mercury(self):
        """Return the value of this quantity as inches-mercury."""
        return self.value

    @property
    def pounds_per_sq_in(self):
        """Return the value of this quantity as pounds-per-square-inch."""
        return self.inches_mercury * 0.4911541996322



"""Unit tests for velocity units."""

from conftest import isclose

from wamu.velocity import (
    FeetPerSecond,
    KilometersPerHour,
    Knots,
    Mach,
    MetersPerSecond,
    MilesPerHour,
)


def test_one_mps():
    """Confirm simple MetersPerSecond conversions."""
    mps = MetersPerSecond(1)

    assert mps == 1.0

    assert float(mps) == 1.0
    assert int(mps) == 1

    assert mps.kilometers_per_hr == 3.6

    assert isclose(mps.miles_per_hr, 2.23694)
    assert isclose(mps.feet_per_sec, 3.28084)
    assert isclose(mps.knots, 1.94384449)

    assert str(mps) == "1 m/s"
    assert repr(mps) == "MetersPerSecond(1)"


def test_one_kph():
    """Confirm simple KilometersPerHour conversions."""
    kph = KilometersPerHour(1)

    assert kph == 1.0

    assert float(kph) == 1.0
    assert int(kph) == 1

    assert isclose(kph.meters_per_sec, 0.27777778)
    assert isclose(kph.miles_per_hr, 0.62137119)
    assert isclose(kph.feet_per_sec, 0.91134442)
    assert isclose(kph.knots, 0.5399568)
    assert isclose(kph.mach, 0.00080985)

    assert str(kph) == "1 km/h"
    assert repr(kph) == "KilometersPerHour(1)"


def test_one_mph():
    """Confirm simple MilesPerHour conversions."""
    mph = MilesPerHour(1)

    assert mph == 1.0

    assert float(mph) == 1.0
    assert int(mph) == 1

    assert isclose(mph.feet_per_sec, 1.46666667)
    assert isclose(mph.meters_per_sec, 0.44704)
    assert isclose(mph.kilometers_per_hr, 1.609344)
    assert isclose(mph.knots, 0.8689762419)
    assert isclose(mph.mach, 0.00130332)

    assert str(mph) == "1 mph"
    assert repr(mph) == "MilesPerHour(1)"


def test_one_fps():
    """Confirm simple FeetPerSecond conversions."""
    fps = FeetPerSecond(1)

    assert fps == 1.0

    assert float(fps) == 1.0
    assert int(fps) == 1

    assert isclose(fps.meters_per_sec, 0.3048)
    assert isclose(fps.kilometers_per_hr, 1.09728)
    assert isclose(fps.miles_per_hr, 0.68181818)
    assert isclose(fps.knots, 0.5924838)

    assert str(fps) == "1 fps"
    assert repr(fps) == "FeetPerSecond(1)"


def test_one_knot():
    """Confirm simple Knot conversions."""
    knot = Knots(1)

    assert knot == 1.0

    assert float(knot) == 1.0
    assert int(knot) == 1

    assert knot.kilometers_per_hr == 1.852
    assert isclose(knot.meters_per_sec, 0.514444)
    assert isclose(knot.miles_per_hr, 1.1507794)
    assert isclose(knot.feet_per_sec, 1.6878099)
    assert isclose(knot.mach, 0.00149984)

    assert str(knot) == "1 kt"
    assert repr(knot) == "Knots(1)"


def test_one_mach():
    """Confirm simple Mach conversions."""
    mach = Mach(1)

    assert mach == 1.0

    assert float(mach) == 1.0
    assert int(mach) == 1

    assert mach.meters_per_sec == 343.0
    assert mach.kilometers_per_hr == 1234.8

    assert isclose(mach.miles_per_hr, 767.269148)
    assert isclose(mach.feet_per_sec, 1125.328084)
    assert isclose(mach.knots, 666.738661)

    assert str(mach) == "1 M"
    assert repr(mach) == "Mach(1)"



"""Utilities for testing."""

import math

COMPARE_TOLERANCE = 1e-5


def isclose(a, b):
    """Determine if the values are "close enough" for testing purposes."""
    return math.isclose(a, b, rel_tol=COMPARE_TOLERANCE)



"""Unit tests for temperature units."""

from wamu.temperature import Celsius, Fahrenheit, Kelvin, Temperature


def assert_is_freezing(temp: Temperature):
    """Assert all values match expected levels for freezing."""
    assert temp.celcius == 0.0
    assert temp.fahrenheit == 32.0
    assert temp.kelvin == 273.15


def test_convert_none():
    """Make sure that conversions with None are also None."""
    tempC = Celsius(None)

    assert tempC.celcius is None
    assert tempC.fahrenheit is None
    assert tempC.kelvin is None


def test_degC_freezing():
    """Confirm Celsius conversions for freezing."""
    tempC = Celsius(0)

    assert tempC == 0.0

    assert int(tempC) == 0
    assert float(tempC) == 0.0

    assert str(tempC) == "0 °C"
    assert repr(tempC) == "Celsius(0)"

    assert_is_freezing(tempC)

    # remove when deprecated property is removed
    assert tempC.degrees_celcius == tempC.celcius


def test_degF_freezing():
    """Confirm Fahrenheit conversions for freezing."""
    tempF = Fahrenheit(32)

    assert tempF == 32.0

    assert int(tempF) == 32
    assert float(tempF) == 32.0

    assert str(tempF) == "32 °F"
    assert repr(tempF) == "Fahrenheit(32)"

    assert_is_freezing(tempF)

    # remove when deprecated property is removed
    assert tempF.degrees_fahrenheit == tempF.fahrenheit


def test_degK_freezing():
    """Confirm Kelvin conversions for freezing."""
    tempK = Kelvin(273.15)

    assert tempK == 273.15

    assert int(tempK) == 273
    assert float(tempK) == 273.15

    assert str(tempK) == "273.15 K"
    assert repr(tempK) == "Kelvin(273.15)"

    assert_is_freezing(tempK)

    # remove when deprecated property is removed
    assert tempK.degrees_kelvin == tempK.kelvin


def test_boiling_temps():
    """Confirm conversions for boiling temperatures."""
    tempF = Fahrenheit(212.0)
    assert tempF.celcius == 100.0
    assert tempF.kelvin == 373.15

    tempC = Celsius(100.0)
    assert tempC.fahrenheit == 212.0
    assert tempF.kelvin == 373.15

    tempK = Kelvin(373.15)
    assert tempK.celcius == 100.0
    assert tempK.fahrenheit == 212.0


def test_mixed_temp_math():
    """Confirm math works with different temperature types."""
    tempC = Celsius(0)
    tempC += Fahrenheit(212)

    assert tempC.celcius == 100.0
    assert tempC.fahrenheit == 212.0

    tempK = Kelvin(0)
    tempK += Celsius(100)

    assert tempK.kelvin == 373.15
    assert tempK.celcius == 100.0


def test_mixed_temp_compare():
    """Confirm comparison works with different temperature types."""
    assert Celsius(0) == Fahrenheit(32)
    assert Celsius(100) > Kelvin(100)

    # make sure numbers can appear on either side of the comparison
    assert 100 >= Kelvin(100)
    assert Kelvin(100) >= 100



"""Working with rate quantities."""

from abc import ABC, abstractproperty

from .quantity import Quantity, UnitSymbol


class RateUnit(UnitSymbol):
    """Symbols for rate units."""

    CENTIMETERS_PER_HOUR = "cm/h"
    CM_PER_HR = "cm/h"
    CMPH = "cm/h"

    MILLIMETERS_PER_HOUR = "mm/h"
    MM_PER_HR = "mm/h"
    MMPH = "mm/h"

    INCHES_PER_HOUR = "in/h"
    IN_PER_HR = "in/h"
    INPH = "in/h"


class Rate(Quantity, ABC):
    """Base for all rate unit types."""

    @abstractproperty
    def centimeters_per_hr(self):
        """Return the value of this quantity as centimeters per hour"""

    @property
    def millimeters_per_hr(self):
        """Return the value of this quantity as millimeters per hour"""
        return self.centimeters_per_hr * 10

    @abstractproperty
    def inches_per_hour(self):
        """Return the value of this quantity as inches per hour"""

    def __call__(self, type):  # noqa: C901
        """Convert this Rate quantity to the given type."""
        if type == CentimetersPerHour:
            return CentimetersPerHour(self.centimeters_per_hr)

        if type == MillimetersPerHour:
            return MillimetersPerHour(self.millimeters_per_hr)

        if type == InchesPerHour:
            return InchesPerHour(self.inches_per_hour)

        raise TypeError(f"Cannot convert to {type}")


class CentimetersPerHour(Rate, symbol=RateUnit.CENTIMETERS_PER_HOUR):
    """A representation of cm / hour."""

    @property
    def centimeters_per_hr(self):
        """Return the value of this quantity as centimeters per second"""
        return self.value

    @property
    def inches_per_hour(self):
        """Return the value of this quantity as inches per hour"""
        return self.centimeters_per_hr / 2.54


class MillimetersPerHour(CentimetersPerHour, symbol=RateUnit.MILLIMETERS_PER_HOUR):
    """A representation of mm / hour."""

    @property
    def centimeters_per_hr(self):
        """Return the value of this quantity as centimeters per second"""
        return self.value * 0.1


class InchesPerHour(Rate, symbol=RateUnit.INCHES_PER_HOUR):
    """A representation of inch / hour."""

    @property
    def centimeters_per_hr(self):
        """Return the value of this quantity as centimeters per second"""
        return self.inches_per_hour * 2.54

    @property
    def inches_per_hour(self):
        """Return the value of this quantity as inches per hour"""
        return self.value



"""Working with velocity quantities."""

from abc import ABC, abstractproperty

from .quantity import Quantity, UnitSymbol


class VelocityUnit(UnitSymbol):
    """Symbols for velocity units."""

    METERS_PER_SECOND = "m/s"
    MPS = "m/s"

    KILOMETERS_PER_HOUR = "km/h"
    KM_PER_HR = "km/h"
    KPH = "km/h"

    MILES_PER_HOUR = "mph"
    MPH = "mph"

    FEET_PER_SECOND = "fps"
    FT_PER_SEC = "fps"
    FPS = "fps"

    KNOTS = "kt"
    KNOT = "kt"

    MACH = "M"


class Velocity(Quantity, ABC):
    """Base for all velocity unit types."""

    @abstractproperty
    def meters_per_sec(self):
        """Return the value of this quantity as meters per second"""

    @property
    def kilometers_per_hr(self):
        """Return the value of this quantity as kilometers per hour"""
        return self.meters_per_sec * 3.6

    @abstractproperty
    def feet_per_sec(self):
        """Return the value of this quantity as feet per second"""

    @property
    def miles_per_hr(self):
        """Return the value of this quantity as miles per hour"""
        return (self.feet_per_sec * 3600) / 5280

    @property
    def knots(self):
        """Return the value of this quantity as knots"""
        return self.kilometers_per_hr / 1.852

    @property
    def mach(self):
        """Return the value of this quantity as mach"""
        return self.meters_per_sec / 343

    def __call__(self, type):  # noqa: C901
        """Convert this Velocity quantity to the given type."""
        if type == MetersPerSecond:
            return MetersPerSecond(self.meters_per_sec)

        if type == KilometersPerHour:
            return KilometersPerHour(self.kilometers_per_hr)

        if type == MilesPerHour:
            return MilesPerHour(self.miles_per_hr)

        if type == FeetPerSecond:
            return FeetPerSecond(self.feet_per_sec)

        if type == Knots:
            return Knots(self.knots)

        if type == Mach:
            return Mach(self.mach)

        raise TypeError(f"Cannot convert {self.__class__} to {type}")


class MetersPerSecond(Velocity, symbol=VelocityUnit.METERS_PER_SECOND):
    """A representation of m/s."""

    @property
    def meters_per_sec(self):
        """Return the value of this quantity as meters per second"""
        return self.value

    @property
    def feet_per_sec(self):
        """Return the value of this quantity as feet per second"""
        return self.meters_per_sec * 3.2808399


class KilometersPerHour(MetersPerSecond, symbol=VelocityUnit.KILOMETERS_PER_HOUR):
    """A representation of km/h."""

    @property
    def meters_per_sec(self):
        """Return the value of this quantity as meters per second"""
        return (self.value * 1000) / 3600


class FeetPerSecond(Velocity, symbol=VelocityUnit.FEET_PER_SECOND):
    """A representation of fps."""

    @property
    def feet_per_sec(self):
        """Return the value of this quantity as feet per second"""
        return self.value

    @property
    def meters_per_sec(self):
        """Return the value of this quantity as meters per second"""
        return self.feet_per_sec / 3.2808399


class MilesPerHour(FeetPerSecond, symbol=VelocityUnit.MILES_PER_HOUR):
    """A representation of mph."""

    @property
    def feet_per_sec(self):
        """Return the value of this quantity as feet per second"""
        return (self.value * 5280) / 3600


class Knots(MetersPerSecond, symbol=VelocityUnit.KNOTS):
    """A representation of knots."""

    @property
    def meters_per_sec(self):
        """Return the value of this quantity as meters per second"""
        return 1.852 * self.value / 3.6


class Mach(MetersPerSecond, symbol=VelocityUnit.MACH):
    """A representation of mach."""

    @property
    def meters_per_sec(self):
        """Return the value of this quantity as meters per second"""
        return self.value * 343



"""Unit tests for velocity units."""

from conftest import isclose

from wamu.rate import CentimetersPerHour, InchesPerHour, MillimetersPerHour


def test_one_cm_per_hour():
    """Confirm simple CentimeterPerHour conversions."""
    cm_per_hr = CentimetersPerHour(1)

    assert cm_per_hr == 1.0

    assert float(cm_per_hr) == 1.0
    assert int(cm_per_hr) == 1

    assert cm_per_hr.millimeters_per_hr == 10.0
    assert isclose(cm_per_hr.inches_per_hour, 0.3937)

    assert str(cm_per_hr) == "1 cm/h"
    assert repr(cm_per_hr) == "CentimetersPerHour(1)"


def test_one_mm_per_hr():
    """Confirm simple MillimeterPerHour conversions."""
    mm_per_hr = MillimetersPerHour(1)

    assert mm_per_hr == 1.0

    assert float(mm_per_hr) == 1.0
    assert int(mm_per_hr) == 1

    assert mm_per_hr.centimeters_per_hr == 0.1
    assert isclose(mm_per_hr.inches_per_hour, 0.03937)

    assert str(mm_per_hr) == "1 mm/h"
    assert repr(mm_per_hr) == "MillimetersPerHour(1)"


def test_one_inch_per_hr():
    """Confirm simple InchPerHour conversions."""
    in_per_hr = InchesPerHour(1)

    assert in_per_hr == 1.0

    assert float(in_per_hr) == 1.0
    assert int(in_per_hr) == 1

    assert in_per_hr.centimeters_per_hr == 2.54
    assert in_per_hr.millimeters_per_hr == 25.4

    assert str(in_per_hr) == "1 in/h"
    assert repr(in_per_hr) == "InchesPerHour(1)"



"""Methods and classes for converting between units."""

from .distance import (
    Centimeter,
    Foot,
    Inch,
    Kilometer,
    Meter,
    Mile,
    Millimeter,
    NauticalMile,
    Yard,
)
from .pressure import Hectopascal, InchesMercury, Pascal
from .rate import InchesPerHour, MillimetersPerHour
from .temperature import Celsius, Fahrenheit, Kelvin
from .velocity import FeetPerSecond, KilometersPerHour, MetersPerSecond, MilesPerHour
from .volume import FluidOunceUS, Gallon, Liter, Milliliter, Pint, Quart
from .weight import Gram, Kilogram, Milligram, Ounce, Pound, Ton

__all__ = [
    "Celsius",
    "Centimeter",
    "FeetPerSecond",
    "Foot",
    "Fahrenheit",
    "FluidOunceUS",
    "Gallon",
    "Gram",
    "Hectopascal",
    "Inch",
    "InchesMercury",
    "InchesPerHour",
    "Kelvin",
    "Kilometer",
    "KilometersPerHour",
    "Kilogram",
    "Liter",
    "Meter",
    "MetersPerSecond",
    "Mile",
    "MilesPerHour",
    "Millimeter",
    "MillimetersPerHour",
    "Milligram",
    "Milliliter",
    "NauticalMile",
    "Ounce",
    "Pascal",
    "Pint",
    "Pound",
    "Quart",
    "Ton",
    "Yard",
]



"""Working with weight quantities."""

from abc import ABC, abstractproperty

from .quantity import Quantity, UnitSymbol


class WeightUnit(UnitSymbol):
    """Symbols for weight units."""

    KILOGRAM = "kg"
    KILOGRAMS = "kg"
    KG = "kg"

    GRAM = "g"
    GRAMS = "g"
    G = "g"

    MILLIGRAM = "mg"
    MILLIGRAMS = "mg"
    MG = "mg"

    POUND = "lb"
    POUNDS = "lb"
    LB = "lb"

    OUNCE = "oz"
    OUNCES = "oz"
    OZ = "oz"

    TON = "ton"
    TONS = "tons"


class Weight(Quantity, ABC):
    """Base for all weight unit types."""

    @abstractproperty
    def kilograms(self):
        """Return the value of this quantity in kilograms."""

    @property
    def grams(self):
        """Return the value of this quantity in grams."""
        return self.kilograms * 1000

    @property
    def milligrams(self):
        """Return the value of this quantity in milligrams."""
        return self.grams * 1000

    @abstractproperty
    def pounds(self):
        """Return the value of this quantity in pounds."""

    @property
    def ounces(self):
        """Return the value of this quantity in ounces."""
        return self.pounds * 16

    @property
    def tons(self):
        """Return the value of this quantity in tons."""
        return self.pounds / 2000

    def __call__(self, type):  # noqa: C901
        """Convert this Weight quantity to the given type."""
        if type == Kilogram:
            return Kilogram(self.kilograms)

        if type == Gram:
            return Gram(self.grams)

        if type == Milligram:
            return Milligram(self.milligrams)

        if type == Pound:
            return Pound(self.pounds)

        if type == Ounce:
            return Ounce(self.ounces)

        if type == Ton:
            return Ton(self.tons)

        raise TypeError(f"Cannot convert to {type}")


class Kilogram(Weight, symbol=WeightUnit.KILOGRAM):
    """A quantity of weight in kilograms."""

    @property
    def kilograms(self):
        """Return the value of this quantity in kilograms."""
        return self.value

    @property
    def pounds(self):
        """Return the value of this quantity in pounds."""
        return self.kilograms * 2.20462262


class Gram(Kilogram, symbol=WeightUnit.GRAM):
    """A quantity of weight in grams."""

    @property
    def kilograms(self):
        """Return the value of this quantity in kilograms."""
        return self.value / 1000


class Milligram(Kilogram, symbol=WeightUnit.MILLIGRAM):
    """A quantity of weight in milligrams."""

    @property
    def kilograms(self):
        """Return the value of this quantity in kilograms."""
        return self.value / 1000000


class Pound(Weight, symbol=WeightUnit.POUND):
    """A quantity of weight in pounds."""

    @property
    def kilograms(self):
        """Return the value of this quantity in kilograms."""
        return self.pounds * 0.45359237

    @property
    def pounds(self):
        """Return the value of this quantity in pounds."""
        return self.value


class Ounce(Pound, symbol=WeightUnit.OUNCE):
    """A quantity of weight in ounces."""

    @property
    def pounds(self):
        """Return the value of this quantity in pounds."""
        return self.value / 16


class Ton(Pound, symbol=WeightUnit.TON):
    """A quantity of weight in standard (short) tons."""

    @property
    def pounds(self):
        """Return the value of this quantity in pounds."""
        return self.value * 2000



"""Base functionality for working with quantities."""

from abc import abstractmethod
from enum import StrEnum


class UnitSymbol(StrEnum):
    """Symbols for all units."""


class Quantity:
    """Base class for all quantities."""

    _unit_symbol = None

    def __init__(self, value: int | float):
        self.value = value

    def __init_subclass__(cls, symbol: UnitSymbol = None, **kwargs):
        """Register all subclasses with their symbols."""
        super().__init_subclass__(**kwargs)

        cls._unit_symbol = symbol

    @abstractmethod
    def __call__(self, type):
        """Convert this Quantity to the given type."""

    def __getattribute__(self, attr):
        """Return the requested attribute of this `Quantity`, or None.

        When callers request an attribute from a `Quantity` object, this method first
        examines the value of this `Quantity` object.  If the value is `None`, this
        method will always return `None` rather than accessing the requested attribute.

        The benefit of this approach is that subclasses do not need to check for `None`
        in each of the conversion methods.  For example, the `Temperature.degC()` method
        does not need to check for `None` since this method will prevent it from being
        called if the value is `None`.

        NOTE - this does not affect the math (dunder) operations on `Quantity` objects.
        """

        # handle special cases to avoid infinite recursion
        if attr == "value" or attr.startswith("_"):
            return object.__getattribute__(self, attr)

        # ... if the contained value is None, return None
        if self.value is None:
            return None

        # otherwise, return the requested attribute
        return object.__getattribute__(self, attr)

    def __float__(self):
        """Return the Quantity value as a `float`."""
        return float(self.value)

    def __int__(self):
        """Return the Quantity value as an `int`."""
        return int(self.value)

    def __add__(self, other):
        """Return the sum of this `Quantity` and `other`."""
        return Quantity(self.value + self._other_value(other))

    def __sub__(self, other):
        """Return the difference of this `Quantity` and `other`."""
        return Quantity(self.value - self._other_value(other))

    def __mul__(self, other):
        """Return the product of this `Quantity` and `other`."""
        return Quantity(self.value * self._other_value(other))

    def __truediv__(self, other):
        """Return the quotient of this `Quantity` and `other`."""
        return Quantity(self.value / self._other_value(other))

    def __iadd__(self, other):
        """Add the given value to this Quantity."""
        self.value += self._other_value(other)

        return self

    def __isub__(self, other):
        """Subtract the given value from this Number."""
        self.value -= self._other_value(other)

        return self

    def __imul__(self, other):
        """Multiply this `Quantity` by `other`."""
        self.value *= self._other_value(other)

        return self

    def __itruediv__(self, other):
        """Divide this `Quantity` by `other`."""
        self.value /= self._other_value(other)

        return self

    def __eq__(self, other):
        """Determine if this `Quantity` value is equal to `other`."""
        return self.value == self._other_value(other)

    def __ne__(self, other):
        """Determine if this property is not equal to the given object."""
        return self.value != self._other_value(other)

    def __le__(self, other):
        """Return `True` if this `Quantity` value is less-than-or-equal-to `other`."""
        return self < other or self == other

    def __lt__(self, other):
        """Return `True` if this `Quantity` value is less-than `other`."""
        return self.value < self._other_value(other)

    def __ge__(self, other):
        """Return `True` if this `Quantity` value is greater-than-or-equal-to `other`."""
        return self > other or self == other

    def __gt__(self, other):
        """Return `True` if this `Quantity` value is greater-than `other`."""
        return self.value > self._other_value(other)

    def __str__(self):
        """Return a human readable string for this quantity."""
        ret = str(self.value)

        if isinstance(self.symbol, UnitSymbol):
            ret += f" {self.symbol.value}"

        return ret

    def __repr__(self):
        """Return a string representation of this quantity."""
        return f"{self.__class__.__name__}({self.value})"

    def _other_value(self, other):
        """Return a float representation of `other`.

        When `other` is a number, this method will simply return the number.

        If `other` is another `Quantity`, this method will convert the value to a
        matching type for this `Quantity` and return the value.

        If `other` is unsupported, this method will raise a `TypeError`.
        """

        if isinstance(other, Quantity):
            # attempt to convert other to a matching type for this Quantity
            if not isinstance(other, self.__class__):
                try:
                    other = other(self.__class__)
                except TypeError as err:
                    raise TypeError("Incompatible types") from err

            return other.value

        return other

    @property
    def symbol(self):
        """Return the unit symbol of this quantity."""
        return self.__class__._unit_symbol



"""Unit tests for weight units."""

from conftest import isclose

from wamu.weight import Gram, Kilogram, Milligram, Ounce, Pound


def test_one_kg():
    """Confirm simple Kilogram conversions."""
    kg = Kilogram(1)

    assert kg == 1.0

    assert float(kg) == 1.0
    assert int(kg) == 1

    assert kg.kilograms == 1.0
    assert kg.grams == 1000.0
    assert kg.milligrams == 1000000.0

    assert isclose(kg.pounds, 2.204622)
    assert isclose(kg.ounces, 35.274)


def test_one_g():
    """Confirm simple Gram conversions."""
    g = Gram(1)

    assert g == 1.0

    assert float(g) == 1.0
    assert int(g) == 1

    assert g.kilograms == 0.001
    assert g.grams == 1.0
    assert g.milligrams == 1000.0

    assert isclose(g.pounds, 0.002204622)
    assert isclose(g.ounces, 0.035274)


def test_one_mg():
    """Confirm simple Milligram conversions."""
    mg = Milligram(1)

    assert mg == 1.0

    assert float(mg) == 1.0
    assert int(mg) == 1

    assert mg.kilograms == 0.000001
    assert mg.grams == 0.001
    assert mg.milligrams == 1.0

    assert isclose(mg.pounds, 0.000002204622)
    assert isclose(mg.ounces, 0.000035274)


def test_one_lb():
    """Confirm simple Pound conversions."""
    lb = Pound(1)

    assert lb == 1.0

    assert float(lb) == 1.0
    assert int(lb) == 1

    assert lb.pounds == 1.0
    assert lb.ounces == 16.0
    assert lb.tons == 0.0005

    assert isclose(lb.kilograms, 0.45359237)


def test_one_oz():
    """Confirm simple Ounce conversions."""
    oz = Ounce(1)

    assert oz == 1.0

    assert float(oz) == 1.0
    assert int(oz) == 1

    assert oz.ounces == 1.0
    assert oz.pounds == 0.0625
    assert oz.tons == 0.00003125

    assert isclose(oz.kilograms, 0.02834952)



"""Unit tests for pressure units."""

from conftest import isclose

from wamu.pressure import (
    Atmosphere,
    Bar,
    Hectopascal,
    InchesMercury,
    Pascal,
    PoundsPerSquareInch,
)


def test_one_pascal():
    """Confirm simple Pascal conversions."""
    pa = Pascal(1)

    assert pa == 1.0

    assert float(pa) == 1.0
    assert int(pa) == 1

    assert pa.bar == 1e-5
    assert isclose(pa.pounds_per_sq_in, 0.0001450377)

    assert str(pa) == "1 Pa"
    assert repr(pa) == "Pascal(1)"


def test_one_hectopascal():
    """Confirm simple Hectopascal conversions."""
    hPa = Hectopascal(1)

    assert hPa == 1.0

    assert float(hPa) == 1.0
    assert int(hPa) == 1

    assert hPa.bar == 1e-3

    assert isclose(hPa.inches_mercury, 0.02953)

    assert str(hPa) == "1 hPa"
    assert repr(hPa) == "Hectopascal(1)"


def test_one_inch_mercury():
    """Confirm simple InchesMercury conversions."""
    inHg = InchesMercury(1)

    assert inHg == 1.0

    assert float(inHg) == 1.0
    assert int(inHg) == 1

    assert isclose(inHg.bar, 0.033864)
    assert isclose(inHg.pounds_per_sq_in, 0.4911542)

    assert str(inHg) == "1 inHg"
    assert repr(inHg) == "InchesMercury(1)"


def test_one_bar():
    """Confirm simple Bar conversions."""
    bar = Bar(1)

    assert bar == 1.0

    assert float(bar) == 1.0
    assert int(bar) == 1

    assert bar.hectopascals == 1000.0
    assert isclose(bar.pounds_per_sq_in, 14.50377)

    assert str(bar) == "1 bar"
    assert repr(bar) == "Bar(1)"


def test_one_atmosphere():
    """Confirm simple Atmosphere conversions."""
    atm = Atmosphere(1)

    assert atm == 1.0

    assert float(atm) == 1.0
    assert int(atm) == 1

    assert isclose(atm.bar, 1.01325)
    assert isclose(atm.pounds_per_sq_in, 14.69595)

    assert str(atm) == "1 atm"
    assert repr(atm) == "Atmosphere(1)"


def test_one_psi():
    """Confirm simple PSI conversions."""
    psi = PoundsPerSquareInch(1)

    assert psi == 1.0

    assert float(psi) == 1.0
    assert int(psi) == 1

    assert isclose(psi.bar, 0.0689476)
    assert isclose(psi.inches_mercury, 2.036020461087)

    assert str(psi) == "1 psi"
    assert repr(psi) == "PoundsPerSquareInch(1)"



"""Unit tests for distance units."""

from conftest import isclose

from wamu.distance import (
    Foot,
    Inch,
    Kilometer,
    LightYear,
    Meter,
    Mile,
    Millimeter,
    NauticalMile,
    Parsec,
    Yard,
)


def test_one_meter():
    """Confirm simple Meter conversions."""
    meter = Meter(1)

    assert meter == 1.0

    assert float(meter) == 1.0
    assert int(meter) == 1

    assert meter.kilometers == 0.001
    assert meter.centimeters == 100.0
    assert meter.millimeters == 1000.0

    assert isclose(meter.feet, 3.28084)
    assert isclose(meter.yards, 1.09361)
    assert isclose(meter.inches, 39.3701)

    assert str(meter) == "1 m"
    assert repr(meter) == "Meter(1)"


def test_one_kilometer():
    """Confirm simple Kilometer conversions."""
    kilo = Kilometer(1)

    assert kilo == 1.0

    assert float(kilo) == 1.0
    assert int(kilo) == 1

    assert kilo.meters == 1000.0

    assert isclose(kilo.miles, 0.62137119)
    assert isclose(kilo.feet, 3280.8399)

    assert str(kilo) == "1 km"
    assert repr(kilo) == "Kilometer(1)"


def test_one_millimeter():
    """Confirm simple Millimeter conversions."""
    mm = Millimeter(1)

    assert mm == 1.0

    assert float(mm) == 1.0
    assert int(mm) == 1

    assert mm.meters == 0.001
    assert mm.centimeters == 0.1
    assert mm.kilometers == 1e-6

    assert isclose(mm.inches, 0.03937)
    assert isclose(mm.feet, 0.00328084)

    assert str(mm) == "1 mm"
    assert repr(mm) == "Millimeter(1)"


def test_one_mile():
    """Confirm simple Mile conversions."""
    mile = Mile(1)

    assert mile == 1.0

    assert float(mile) == 1.0
    assert int(mile) == 1

    assert mile.feet == 5280.0
    assert mile.yards == 1760.0
    assert mile.inches == 63360.0

    assert mile.meters == 1609.344
    assert mile.kilometers == 1.609344

    assert str(mile) == "1 mi"
    assert repr(mile) == "Mile(1)"


def test_one_yard():
    """Confirm simple Yard conversions."""
    yard = Yard(1)

    assert yard == 1.0

    assert float(yard) == 1.0
    assert int(yard) == 1

    assert yard.inches == 36.0
    assert isclose(yard.miles, 0.00056818)

    assert isclose(yard.meters, 0.9144)
    assert isclose(yard.centimeters, 91.44)

    assert str(yard) == "1 yd"
    assert repr(yard) == "Yard(1)"


def test_one_foot():
    """Confirm simple Feet conversions."""
    foot = Foot(1)

    assert foot == 1.0

    assert float(foot) == 1.0
    assert int(foot) == 1

    assert foot.inches == 12.0
    assert foot.centimeters == 30.48

    assert str(foot) == "1 ft"
    assert repr(foot) == "Foot(1)"


def test_one_inch():
    """Confirm simple Inch conversions."""
    inch = Inch(1)

    assert inch == 1.0

    assert float(inch) == 1.0
    assert int(inch) == 1

    assert isclose(inch.feet, 0.08333333)
    assert isclose(inch.yards, 0.02777778)
    assert isclose(inch.miles, 1.5782828e-5)

    assert inch.centimeters == 2.54
    assert inch.millimeters == 25.4

    assert str(inch) == "1 in"
    assert repr(inch) == "Inch(1)"


def test_one_nautical_mile():
    """Confirm simple Nautical Mile conversions."""
    nm = NauticalMile(1)

    assert nm == 1.0

    assert float(nm) == 1.0
    assert int(nm) == 1

    assert nm.meters == 1852.0
    assert nm.kilometers == 1.852

    assert isclose(nm.feet, 6076.1155)
    assert isclose(nm.yards, 2025.3718)
    assert isclose(nm.miles, 1.1507794)

    assert str(nm) == "1 NM"
    assert repr(nm) == "NauticalMile(1)"


def test_one_parsec():
    """Confirm simple Parsec conversions."""
    pc = Parsec(1)

    assert pc == 1.0

    assert float(pc) == 1.0
    assert int(pc) == 1

    assert isclose(pc.miles, 19173513289161)
    assert isclose(pc.kilometers, 30856778570831)
    assert isclose(pc.light_years, 3.261563777)

    assert str(pc) == "1 pc"
    assert repr(pc) == "Parsec(1)"


def test_one_light_year():
    """Confirm simple Light Year conversions."""
    ly = LightYear(1)

    assert ly == 1.0

    assert float(ly) == 1.0
    assert int(ly) == 1

    assert ly.kilometers == 9460730472580.8

    assert isclose(ly.miles, 5878625373183)
    assert isclose(ly.parsecs, 0.30660139)

    assert str(ly) == "1 ly"
    assert repr(ly) == "LightYear(1)"



"""Unit tests for the `units` module."""

import pytest

from wamu.quantity import Quantity


def test_basic_none():
    """Check that quantities of None behave as expected."""
    qty = Quantity(None)

    assert qty.value is None
    assert qty == None  # noqa: E711

    assert str(qty) == "None"
    assert repr(qty) == "Quantity(None)"


def test_none_math():
    """Check that math with None behaves as expected."""
    qty = Quantity(None)

    with pytest.raises(TypeError):
        qty + 1

    with pytest.raises(TypeError):
        qty *= 1

    with pytest.raises(TypeError):
        float(qty)

    with pytest.raises(TypeError):
        Quantity(1) - None


def test_qty_zero():
    """Check that Quantity(0) behaves as expected."""
    qty = Quantity(0)

    assert qty.value == 0
    assert qty == 0
    assert qty == 0.0
    assert qty == Quantity(0)

    assert qty + 0 == 0
    assert qty - 0 == 0
    assert qty * 1 == 0
    assert qty / 1 == 0


def test_qty_iadd():
    """Check that Quantity.__iadd__ works as expected."""
    qty_a = Quantity(1)

    # using a scalar
    qty_a += 1

    assert qty_a == 2

    # using a Quantity object
    qty_b = Quantity(1)

    qty_a += qty_b

    assert qty_a == 3
    assert qty_b == 1


def test_qty_isub():
    """Check that Quantity.__isub__ works as expected."""
    qty_a = Quantity(10)

    # using a scalar
    qty_a -= 1

    assert qty_a == 9

    # using a Quantity object
    qty_b = Quantity(4)

    qty_a -= qty_b

    assert qty_a == 5
    assert qty_b == 4


def test_qty_imul():
    """Check that Quantity.__imul__ works as expected."""
    qty_a = Quantity(2)

    # using a scalar
    qty_a *= 2

    assert qty_a == 4

    # using a Quantity object
    qty_b = Quantity(3)

    qty_a *= qty_b

    assert qty_a == 12
    assert qty_b == 3


def test_qty_idiv():
    """Check that Quantity.__itruediv__ works as expected."""
    qty_a = Quantity(10)

    # using a scalar
    qty_a /= 2

    assert qty_a == 5

    # using a Quantity object
    qty_b = Quantity(2)

    qty_a /= qty_b

    assert qty_a == 2.5
    assert qty_b == 2


def test_qty_equality():
    """Check that Quantity.__eq__ works as expected."""
    qty_a = Quantity(1)

    assert qty_a == 1
    assert qty_a == 1.0
    assert qty_a == Quantity(1)

    assert qty_a != -1
    assert qty_a != 0.0
    assert qty_a != Quantity(-1)


def test_qty_less_than():
    """Check that Quantity.__lt__ and Quantity.__le__ work as expected."""
    qty_a = Quantity(1)

    # using scalars
    assert qty_a < 2.0
    assert qty_a <= 1.0 and qty_a <= 2.0
    assert 2.0 > qty_a

    # using Quantity objects
    assert qty_a < Quantity(2)
    assert qty_a <= Quantity(1) and qty_a <= Quantity(2)


def test_qty_greater_than():
    """Check that Quantity.__gt__ and Quantity.__ge__ work as expected."""
    qty_a = Quantity(1)

    # using scalars
    assert qty_a > 0.0
    assert qty_a >= 1.0 and qty_a >= 0.0
    assert 0.0 < qty_a

    # using Quantity objects
    assert qty_a > Quantity(0)
    assert qty_a >= Quantity(1) and qty_a >= Quantity(0)



"""Unit tests for volume units."""

from conftest import isclose

from wamu.volume import FluidOunceUS, Gallon, Liter, Milliliter, Pint, Quart


def test_one_liter():
    """Confirm simple Liter conversions."""
    liter = Liter(1)

    assert liter == 1.0

    assert float(liter) == 1.0
    assert int(liter) == 1

    assert liter.liters == 1.0
    assert liter.milliliters == 1000.0

    assert isclose(liter.gallons, 0.2641720524)
    assert isclose(liter.pints, 2.1133764189)
    assert isclose(liter.quarts, 1.0566882094)
    assert isclose(liter.fl_ounces_us, 33.814022702)

    assert str(liter) == "1 L"
    assert repr(liter) == "Liter(1)"


def test_one_ml():
    """Confirm simple Milliliter conversions."""
    ml = Milliliter(1)

    assert ml == 1.0

    assert float(ml) == 1.0
    assert int(ml) == 1

    assert ml.milliliters == 1.0
    assert ml.liters == 0.001

    assert isclose(ml.fl_ounces_us, 0.0338140227)

    assert str(ml) == "1 mL"
    assert repr(ml) == "Milliliter(1)"


def test_one_gal():
    """Confirm simple Gallon conversions."""
    gal = Gallon(1)

    assert gal == 1.0

    assert float(gal) == 1.0
    assert int(gal) == 1

    assert gal.gallons == 1.0
    assert gal.pints == 8.0
    assert gal.quarts == 4.0
    assert gal.fl_ounces_us == 128.0

    assert isclose(gal.liters, 3.78541178)
    assert isclose(gal.milliliters, 3785.41178)

    assert str(gal) == "1 gal"
    assert repr(gal) == "Gallon(1)"


def test_one_pint():
    """Confirm simple Pint conversions."""
    pint = Pint(1)

    assert pint == 1.0

    assert float(pint) == 1.0
    assert int(pint) == 1

    assert pint.pints == 1.0
    assert pint.quarts == 0.5
    assert pint.fl_ounces_us == 16.0

    assert isclose(pint.liters, 0.473176473)
    assert isclose(pint.milliliters, 473.176473)

    assert str(pint) == "1 pt"
    assert repr(pint) == "Pint(1)"


def test_one_quart():
    """Confirm simple Quart conversions."""
    quart = Quart(1)

    assert quart == 1.0

    assert float(quart) == 1.0
    assert int(quart) == 1

    assert quart.quarts == 1.0
    assert quart.fl_ounces_us == 32.0

    assert isclose(quart.liters, 0.946352946)
    assert isclose(quart.milliliters, 946.352946)

    assert str(quart) == "1 qt"
    assert repr(quart) == "Quart(1)"


def test_one_us_oz():
    """Confirm simple US fluid ounce conversions."""
    fl_oz = FluidOunceUS(1)

    assert fl_oz == 1.0

    assert float(fl_oz) == 1.0
    assert int(fl_oz) == 1

    assert fl_oz.pints == 0.0625
    assert fl_oz.quarts == 0.03125
    assert fl_oz.gallons == 0.0078125

    assert isclose(fl_oz.liters, 0.0295735296)
    assert isclose(fl_oz.milliliters, 29.5735295625)

    assert str(fl_oz) == "1 fl oz"
    assert repr(fl_oz) == "FluidOunceUS(1)"



"""Working with volume quantities."""

from abc import ABC, abstractproperty

from .quantity import Quantity, UnitSymbol


class VolumeUnit(UnitSymbol):
    """Symbols for volume units."""

    LITER = "L"
    LITERS = "L"
    L = "L"

    MILLILITER = "mL"
    MILLILITERS = "mL"
    ML = "mL"

    GALLON = "gal"
    GALLONS = "gal"
    GAL = "gal"

    PINT = "pt"
    PINTS = "pt"
    PT = "pt"

    QUART = "qt"
    QUARTS = "qt"
    QT = "qt"

    US_FLUID_OUNCE = "fl oz"
    US_FLUID_OUNCES = "fl oz"
    US_FL_OZ = "fl oz"
    US_OZ = "fl oz"


class Volume(Quantity, ABC):
    """Base for all volume unit types."""

    @abstractproperty
    def liters(self):
        """Return the value of this quantity in liters."""

    @property
    def milliliters(self):
        """Return the value of this quantity in milliliters."""
        return self.liters * 1000

    @abstractproperty
    def gallons(self):
        """Return the value of this quantity in gallons."""

    @property
    def pints(self):
        """Return the value of this quantity in pints."""
        return self.gallons * 8

    @property
    def quarts(self):
        """Return the value of this quantity in quarts."""
        return self.gallons * 4

    @property
    def fl_ounces_us(self):
        """Return the value of this quantity in US fluid ounces."""
        return self.gallons * 128

    def __call__(self, type):  # noqa: C901
        """Convert this Volume quantity to the given type."""
        if type == Liter:
            return Liter(self.liters)

        if type == Milliliter:
            return Milliliter(self.milliliters)

        if type == Gallon:
            return Gallon(self.gallons)

        if type == Pint:
            return Pint(self.pints)

        if type == Quart:
            return Quart(self.quarts)

        if type == FluidOunceUS:
            return FluidOunceUS(self.fl_ounces_us)

        raise TypeError(f"Cannot convert to {type}")


class Liter(Volume, symbol=VolumeUnit.LITER):
    """A quantity of volume in liters."""

    @property
    def liters(self):
        """Return the value of this quantity in liters."""
        return self.value

    @property
    def gallons(self):
        """Return the value of this quantity in gallons."""
        return self.liters * 0.2641720524


class Milliliter(Liter, symbol=VolumeUnit.MILLILITER):
    """A quantity of volume in milliliters."""

    @property
    def liters(self):
        """Return the value of this quantity in liters."""
        return self.value / 1000


class Gallon(Liter, symbol=VolumeUnit.GALLON):
    """A quantity of volume in gallons."""

    @property
    def liters(self):
        """Return the value of this quantity in liters."""
        return self.gallons * 3.785411784

    @property
    def gallons(self):
        """Return the value of this quantity in gallons."""
        return self.value


class Pint(Gallon, symbol=VolumeUnit.PINT):
    """A quantity of volume in pints."""

    @property
    def gallons(self):
        """Return the value of this quantity in gallons."""
        return self.value / 8


class Quart(Gallon, symbol=VolumeUnit.QUART):
    """A quantity of volume in quarts."""

    @property
    def gallons(self):
        """Return the value of this quantity in gallons."""
        return self.value / 4


class FluidOunceUS(Gallon, symbol=VolumeUnit.US_FLUID_OUNCE):
    """A quantity of volume in fluid ounces (US)."""

    @property
    def gallons(self):
        """Return the value of this quantity in gallons."""
        return self.value / 128



"""Working with temperature quantities."""

from abc import ABC, abstractproperty

from deprecation import deprecated

from .quantity import Quantity, UnitSymbol


class TemperatureUnit(UnitSymbol):
    """Symbols for temperature units."""

    CELCIUS = "°C"
    DEGREES_CELCIUS = "°C"
    C = "°C"
    DEGREES_C = "°C"

    FAHRENHEIT = "°F"
    DEGREES_FAHRENHEIT = "°F"
    F = "°F"
    DEGREES_F = "°F"

    KELVIN = "K"
    DEGREES_KELVIN = "K"
    K = "K"
    DEGREES_K = "K"


class Temperature(Quantity, ABC):
    """Base for all temperature unit types."""

    @abstractproperty
    def celcius(self):
        """Return the value of this quantity as Celsius."""

    @property
    @deprecated(
        deprecated_in="0.1",
        removed_in="1.0",
        details="Use the 'celcius' property instead",
    )
    def degrees_celcius(self):
        """Return the value of this quantity as Celsius."""
        return self.celcius

    @abstractproperty
    def fahrenheit(self):
        """Return the value of this quantity as Fahrenheit."""

    @property
    @deprecated(
        deprecated_in="0.1",
        removed_in="1.0",
        details="Use the 'fahrenheit' property instead",
    )
    def degrees_fahrenheit(self):
        """Return the value of this quantity as Fahrenheit."""
        return self.fahrenheit

    @property
    def kelvin(self):
        """Return the value of this quantity as Kelvin."""
        return self.celcius + 273.15

    @property
    @deprecated(
        deprecated_in="0.1",
        removed_in="1.0",
        details="Use the 'kelvin' property instead",
    )
    def degrees_kelvin(self):
        """Return the value of this quantity as Kelvin."""
        return self.kelvin

    def __call__(self, type):  # noqa: C901
        """Convert this Temperature quantity to the given type."""
        if type == Celsius:
            return Celsius(self.celcius)

        if type == Fahrenheit:
            return Fahrenheit(self.fahrenheit)

        if type == Kelvin:
            return Kelvin(self.kelvin)

        raise TypeError(f"Cannot convert to {type}")


class Celsius(Temperature, symbol=TemperatureUnit.CELCIUS):
    """A representation of Celsius quantities."""

    @property
    def celcius(self):
        """Return the value of this quantity as Celsius."""
        return self.value

    @property
    def fahrenheit(self):
        """Return the value of this quantity as Fahrenheit."""
        return (self.celcius * 1.8) + 32


class Fahrenheit(Temperature, symbol=TemperatureUnit.FAHRENHEIT):
    """A representation of Fahrenheit quantities."""

    @property
    def celcius(self):
        """Return the value of this quantity as Celsius."""
        return (self.fahrenheit - 32) / 1.8

    @property
    def fahrenheit(self):
        """Return the value of this quantity as Fahrenheit."""
        return self.value


class Kelvin(Celsius, symbol=TemperatureUnit.KELVIN):
    """A representation of Kelvin quantities."""

    @property
    def celcius(self):
        """Return the value of this quantity as Celsius."""
        return self.value - 273.15
